/*
 * $Id: hdn_common.c,v 1.11 2003/02/04 06:54:13 xvr Exp $
 * Created: 08/21/2002
 *
 * xvr (c) 2002
 * xvr@crazyboy.com
 */

#include "hdn_common.h"

/*
 * xor two streams together
 */
void hdn_xor (void *dest, void *src, int n)
{
    int i;

    for (i = 0; i < n; i++)
	*((char *)dest + i) ^= *((char *)src + i);
}

/*
 * returns the number of bits needed to encode n
 */
int hdn_log2 (long long n)
{
    int i = sizeof (n) * 8;

    while (--i)
    {
	if ((n >> i) & 1)
	    return i;
    }

    return 0;
}

/*
 * seeds the random number generator
 */
void hdn_srandom (char *pass)
{
    unsigned long seed = 123;
    unsigned char *digest;
    
    if ((digest = hdn_hash (pass)))
    {
	/*
	 * gets the first 4 bytes of the hash to use as a seed.
	 */
	memcpy (&seed, digest, sizeof seed);
    }
    else
    {
	fprintf (stderr, "Error: invalid pass. Using default seed instead.\n");
    }

    srandom (seed);
}

/*
 * skips upto 'max' instructions
 */
static int skip_insn (unsigned long max)
{
    static unsigned long left = 0;

    if (!left)
    {
	left = random () % HDN_MAX_SKIP_INSNS;
	return (left != 0);
    }

    left--;

    return (left != 0);
}

#define INS_VALID 0x01
#define INS_USED  0x02


static void init_next_address_ds (struct hdn_host_section *host_sections)
{
    struct instr insn;
    char *addr;
    long long int host_curr_pos = 0;
    struct hdn_host_section *my_host_section;

    my_host_section = host_sections;

    /*
     * tag all valid instructions as such
     */
    while (my_host_section)
    {
	bzero (&insn, sizeof insn);
	addr = my_host_section->data.content + host_curr_pos;
	disassemble_address (addr, &insn);

	/*
	 * if disas is unbounded, goto next section
	 */
	if ((host_curr_pos + insn.size) > my_host_section->data.sz)
	{
	    my_host_section = my_host_section->next;
	    host_curr_pos   = 0;
	    continue;
	}

	/*
	 * tag insn valid if need be
	 * (set first byte only)
	 */
	if (hdn_is_valid_insn (&insn, addr))
	{
	    addr[my_host_section->data.sz] = INS_VALID;
	}

	/*
	 * move to next instruction
	 */
	if (insn.size) host_curr_pos += insn.size;
	else           host_curr_pos++;
    }
}

static long long int first_valid_insn (long long int pos, struct hdn_host_section *sec)
{
    char *addr;

    addr = sec->data.content + sec->data.sz;

    for (; pos < sec->data.sz; pos++)
	if (addr[pos] == INS_VALID)
	    break;

    return pos;
}

/*
 * returns the next address we should embed/decode from the code sections.
 */
char *hdn_get_next_address (struct hdn_host_section *host_sections)
{
    char *addr;
    int ret = 0;
    int next_insn = 0;
    static int first_time = 1;
    static long long int host_curr_pos = 0;
    static struct hdn_host_section *my_host_section = NULL;

    /*
     * init
     */
    if (first_time)
    {
	init_next_address_ds (host_sections);

	/*
	 * go back to the beginning
	 */
	my_host_section = host_sections;
	first_time      = 0;
	goto next_insn;
    }

    while (1)
    {
	/*
	 * no more sections to embed in
	 */
	if (!my_host_section) 
	{
	    //we have any sections to go back to?
	    if (!host_sections) 
		return NULL;

	    host_curr_pos = first_valid_insn (0, host_sections);

	    //if there is an unused instruction in the current section, try it.
	    if (host_curr_pos < host_sections->data.sz)
	    {
		my_host_section = host_sections;
	    }
	    else
	    {
		/*
		 * no more instructions in this section, so discard it.
		 * try to find instructions in the next sections instead.
		 */
		host_sections   = host_sections->next;
		my_host_section = NULL;
		host_curr_pos   = 0;
	    }

	    continue;
	}

	/*
	 * random walk
	 */ 
	if (skip_insn (100))
	{
	    host_curr_pos++;
	    goto next_insn;
	}

	/*
	 * mark used instruction
	 */
	addr = my_host_section->data.content + host_curr_pos;
	addr[my_host_section->data.sz] = INS_USED;

	/*
	 * move onto next host position, and return the current address
	 */
	host_curr_pos++;
	ret = 1;
	goto next_insn;

      next_section:
	my_host_section = my_host_section->next;
	host_curr_pos = 0;
	next_insn = 0;

      next_insn:
	if (my_host_section)
	{
	    host_curr_pos = first_valid_insn (host_curr_pos, my_host_section);
	
	    //not found
	    if (host_curr_pos >= my_host_section->data.sz)
		goto next_section;
	}

	if (ret) return addr;
    }
}

#if (defined(__CYGWIN32__) || defined(_Windows) || defined(_WIN32))
int hdn_is_valid_exe (unsigned char *bin)
{
    size_t pos = 0;
    IMAGE_DOS_HEADER      dos_hdr;
    DWORD                 nt_sig;

    /* read and validate the dos header signature */
    memcpy (&dos_hdr, bin, sizeof (IMAGE_DOS_HEADER));
    if (dos_hdr.e_magic != IMAGE_DOS_SIGNATURE)
    {
	fprintf (stderr, "Host file is not in PE format: DOS sig mismatch\n");
	return 0;
    }

    /* get NT header */
    pos = dos_hdr.e_lfanew;
    memcpy (&nt_sig, bin + pos, sizeof (IMAGE_NT_SIGNATURE));
    pos += sizeof (IMAGE_NT_SIGNATURE);

    if (nt_sig != IMAGE_NT_SIGNATURE)
    {
	fprintf (stderr, "Host file is not in PE format: NT sig mismatch\n");
	return 0;
    }

    return 1;
}
#else
/*
 * returns true if we're dealing with an ELF file.
 */
int hdn_is_valid_exe (unsigned char *bin) 
{
    char magic[] = {ELFMAG0, ELFMAG1, ELFMAG2, ELFMAG3};
    Elf32_Ehdr *ehdr = (Elf32_Ehdr *)bin;

    /* make sure this is in elf format */
    if (memcmp (ehdr->e_ident, magic, sizeof magic))
    {
	fprintf (stderr, "Host file is not in ELF format\n");
	return 0;
    }

    /* we only support elf32 lsb */
    if (ehdr->e_ident[EI_CLASS] != ELFCLASS32 ||
	ehdr->e_ident[EI_DATA ] != ELFDATA2LSB)
    {
	fprintf (stderr, "Host file is not in elf32 lsb format\n");
	return 0;
    }

    /* make sure this is indeed an executable */
    if (ehdr->e_type != ET_EXEC)
    {
	fprintf (stderr, "Host file is not an executable\n");
	return 0;
    }

    if (!ehdr->e_phoff)
    {
	fprintf (stderr, "Host file has no ELF program header\n");
	return 0;
    }

    if (!ehdr->e_shoff)
    {
	fprintf (stderr, "Host file has no ELF section header\n");
	return 0;
    }

    if (ehdr->e_shstrndx == SHN_UNDEF)
    {
	fprintf (stderr, "Host file has no section name string table\n");
	return 0;
    }

    //valid
    return 1;
}
#endif

/*
 * Adds a section to the end of the embeddable sections linked list.
 */
static void hdn_insert_section (struct hdn_host_section **base,
				struct hdn_host_section *new)
{
    struct hdn_host_section *tmp;

    if (!new || !base) return;

    if (!(*base))
    {
	(*base) = new;
	return;
    }

    tmp = (*base);
    while (tmp->next) 
        tmp = tmp->next;

    tmp->next = new;
}

/*
 * Returns the code locations in the exe file as a linked list.  These
 * sections are those that we want to look into for embeddable data.
 */
#if (defined(__CYGWIN32__) || defined(_Windows) || defined(_WIN32))
struct hdn_host_section *hdn_get_code_locs (unsigned char *file_start)
{
    int i;
    struct hdn_host_section *sections = NULL;
    struct hdn_host_section *newsectn = NULL;
    size_t pos = 0;
    IMAGE_DOS_HEADER      dos_hdr;
    IMAGE_FILE_HEADER     file_hdr;
    IMAGE_SECTION_HEADER  sect_hdr;

    /* read the dos header */
    memcpy (&dos_hdr, file_start, sizeof (IMAGE_DOS_HEADER));
    pos = dos_hdr.e_lfanew;

    /* bypass NT header */
    pos += sizeof (IMAGE_NT_SIGNATURE);

    /* get file and bypass optional header */
    memcpy (&file_hdr, file_start + pos, sizeof (IMAGE_FILE_HEADER));
    pos += sizeof (IMAGE_FILE_HEADER);
    pos += sizeof (IMAGE_OPTIONAL_HEADER);

    /*
     * find code sections 
     */
    for (i = 0; i < file_hdr.NumberOfSections; i++)
    {
	memcpy (&sect_hdr, file_start + pos, sizeof (IMAGE_SECTION_HEADER));
	pos += sizeof (IMAGE_SECTION_HEADER);

	if (sect_hdr.Characteristics & IMAGE_SCN_CNT_CODE &&
	    sect_hdr.Characteristics & IMAGE_SCN_MEM_EXECUTE)
	{
	    /*
	     * allocate an extra element in the array
	     *
	     * Note: the latter half of the allocated space is used to
	     * mark which bytes of the first half have already been
	     * used.
	     */
	    newsectn = malloc (sizeof (struct hdn_host_section) + 
			       2 * sect_hdr.SizeOfRawData);

	    /*
	     * copy in the section info and data
	     */
	    newsectn->next    = NULL;
	    newsectn->offset  = sect_hdr.VirtualAddress;
	    newsectn->data.sz = sect_hdr.SizeOfRawData;

	    bzero (newsectn->data.content, 2 * sect_hdr.SizeOfRawData);
	    memcpy (newsectn->data.content,
		    file_start + newsectn->offset,
		    newsectn->data.sz);

	    hdn_insert_section (&sections, newsectn);

#if 0
	    fprintf (stderr, "%s, %d bytes\n",
		     sect_hdr.Name,
		     sect_hdr.SizeOfRawData);
#endif	    
	}
    }

    if (!sections)
        fprintf (stderr, "couldn't find a code section in the PE file\n");

    return sections;
}
#else //ELF
struct hdn_host_section *hdn_get_code_locs (unsigned char *file_start)
{
    int i;
    struct hdn_host_section *sections = NULL;
    struct hdn_host_section *newsectn = NULL;
    Elf32_Ehdr *ehdr = (Elf32_Ehdr *) file_start;
    Elf32_Shdr *shdr = (Elf32_Shdr *)(file_start + ehdr->e_shoff);

    /*
     * find code sections
     */
    for (i=0; i < ehdr->e_shnum; i++)
    {
	if (shdr[i].sh_type == SHT_PROGBITS &&
	    shdr[i].sh_flags & SHF_ALLOC    &&
	    shdr[i].sh_flags & SHF_EXECINSTR)
	{
	    /*
	     * allocate an extra element in the array
	     *
	     * Note: the latter half of the allocated space is used to
	     * mark which bytes of the first half have already been
	     * used.
	     */
	    newsectn = malloc (sizeof (struct hdn_host_section) + 
			       2 * shdr[i].sh_size);

	    /*
	     * copy in the section info and data
	     */
	    newsectn->next    = NULL;
	    newsectn->offset  = shdr[i].sh_offset;
	    newsectn->data.sz = shdr[i].sh_size;

	    bzero (newsectn->data.content, 2 * shdr[i].sh_size);
	    memcpy (newsectn->data.content,
		    file_start + newsectn->offset,
		    newsectn->data.sz);

	    hdn_insert_section (&sections, newsectn);

#if 0  
	    fprintf (stderr, "%s, %d bytes\n",
		     file_start +
		     shdr[ehdr->e_shstrndx].sh_offset + shdr[i].sh_name,
		     shdr[i].sh_size);
#endif
	}
    }

    if (!sections)
	fprintf (stderr, "couldn't find a code section in the ELF file\n");

    return sections;
}
#endif

/*
 * reads into hdn_data as much as possible from a file descriptor
 */
struct hdn_data *hdn_fdread (int fd)
{
    struct stat sta;
    struct hdn_data *data;

    if (fstat (fd, &sta) < 0)
    {
	perror ("fstat");
	return NULL;
    }

    if (!(data = malloc (sizeof (struct hdn_data) + sta.st_size)))
    {
	perror ("malloc");
	return NULL;
    }
    
    //fill out the data structure
    data->sz = sta.st_size;

    if (read (fd, data->content, data->sz) < 0)
    {
	perror ("read");
	return NULL;
    }

    close (fd);
    return data;
}

/*
 * same as above, but from a filename
 */
struct hdn_data *hdn_fileread (char *filename)
{
    int fd;

    if ( (fd = open (filename, O_RDONLY)) < 0)
    {
	perror ("open");
	return NULL;
    }

    return hdn_fdread (fd);
}

/*
 * writes to a file descriptor
 */
void hdn_fdwrite (int fd, struct hdn_data *data)
{
    if (write (fd, data->content, data->sz) < 0)
    {
	perror ("write");
	exit (1);
    }

    close (fd);
}

/*
 * writes to a filename
 */
void hdn_filewrite (char *filename, struct hdn_data *data)
{
    int fd;

    if ( (fd = open (filename, O_CREAT | O_WRONLY | O_TRUNC)) < 0)
    {
	perror ("open");
	exit (1);
    }

    hdn_fdwrite (fd, data);
}

