/*
 * $Id: hdn_embed.c,v 1.15 2003/02/04 06:54:13 xvr Exp $
 * Created: 08/21/2002
 * 
 * xvr (c) 2002-2003
 * xvr@crazyboy.com
 */

#include "hdn_embed.h"

static void hdn_usage (void)
{
    fprintf (stderr,
	     "Usage:\n"
	     "\thydan host_file [message_file]\n\n"
	     "Takes in a binary executable (host_filename) and stores the\n"
	     "message in message_filename inside of it.\n"
	     "Takes input from stdin when message_file is not specified.\n"
	     "Resulting application is output on stdout.\n\n"
	     "Ex: ./hydan /bin/ls <msg> ls.stegged\n\n"
	     "Use hydan-decode to retrieve the hidden message\n");
    exit (1);
}

/*
 * returns the number of successfully embedded bytes
 */
static long long int hdn_embed_message (struct hdn_host_section *host_sections,
					struct hdn_data *mesg_data,
					unsigned char *key)
{
    static long long int mesg_curr_pos = 0;
    static char  mesg_curr_bit = 0;
    char *addr = NULL;
    struct instr insn;
    long long int mesg_sz = mesg_data->sz;

    /*
     * while there are still sections to embed into, and we still have
     * data to embed.
     */
    while ((addr = hdn_get_next_address (host_sections)) &&
	   (mesg_curr_pos < mesg_sz))
    {
	bzero (&insn, sizeof insn);
	disassemble_address (addr, &insn);
	
	/*
	 * can we embed something into this instruction?
	 */
	if (hdn_is_valid_insn (&insn, addr))
	{
#if 0
	    struct instr insn2;
	    int old = mesg_curr_bit;
#endif
	    /*
	     * embed a bit (or more) into this instruction, and return
	     * the number of bits embedded.
	     */
	    mesg_curr_bit += 
		hdn_flip_insn (&insn, addr,
			       mesg_data->content + mesg_curr_pos,
			       mesg_curr_bit);
	    
#if 0
	    disassemble_address (addr, &insn2);
 
	    if (mesg_curr_bit - old)
	    fprintf (stderr, "%02x%02x: %s %s, %s -> %s %s, %s\t%d\t%d\n",
	     (unsigned char)*(addr),
	     (unsigned char)*(addr + 1),
		     insn.mnemonic, insn.dest, insn.src,
		     insn2.mnemonic, insn2.dest, insn2.src, 
		     mesg_curr_bit - old, mesg_curr_pos);
#endif

	    /*
	     * increment mesg position and wrap byte counter around if
	     * necessary.
	     */
	    if (mesg_curr_bit >= 8) mesg_curr_pos++;
	    mesg_curr_bit %= 8;
	}
    }

    return mesg_curr_pos;
}

int hdn_embed_main (int argc, char **argv)
{
    long long int bytes_embedded, code_sz = 0;
    unsigned char   *password;
    struct hdn_data *host_data = 0, *mesg_data = 0;
    struct hdn_host_section *code_sections, *tmp_sections;

    //for stats
    unsigned int num_bits = 0;

    /*
     * get message data
     */
    if      (argc == 2) mesg_data = hdn_fdread (STDIN_FILENO);
    else if (argc == 3) mesg_data = hdn_fileread (argv[2]);
    else                hdn_usage ();

    /*
     * get password and encrypt message data
     */
    password = getpass ("Password: ");
    hdn_srandom (password);
    hdn_encrypt (&mesg_data, password);
    
    /*
     * read in application data
     */
    if (!(host_data = hdn_fileread (argv[1])))
	goto out;

    disassemble_init (0, INTEL_SYNTAX);

    /*
     * get the text segment
     */
    if (!hdn_is_valid_exe (host_data->content))
	goto out;

    if (!(code_sections = hdn_get_code_locs (host_data->content)))
    {
	fprintf (stderr, "Error extracting .text segment from host file\n");
	goto out;
    }

    /*
     * count the number of embeddable instructions in all of the code
     * sections
     */
    tmp_sections = code_sections;
    while (tmp_sections)
    {
	code_sz += tmp_sections->data.sz;

	hdn_stats_embeddable_bits (&tmp_sections->data, &num_bits);
	tmp_sections = tmp_sections->next;
    }

    /* 
     * make sure stego is possible 
     */
    if (mesg_data->sz * 8  >  num_bits)
    {
	fprintf (stderr, 
		 "Not enough place in host application to hide message.\n"  
		 "Can only hide %u bytes in this application "
		 "(needed: %u bytes).\n"
		 "So choose a different host, or make your message smaller!\n",
		 (unsigned int)num_bits/8, 
		 (unsigned int)mesg_data->sz);
	goto out;
    }

    /*
     * actually go thru and embed the message into the host proggie
     */
    bytes_embedded = hdn_embed_message (code_sections, mesg_data, password);

    /*
     * patch modified code sections
     */
    tmp_sections = code_sections;
    while (tmp_sections)
    {
	memcpy (host_data->content + tmp_sections->offset,
		tmp_sections->data.content, 
		tmp_sections->data.sz);

	tmp_sections = tmp_sections->next;
    }

    /*
     * save new app
     */
    hdn_fdwrite (STDOUT_FILENO, host_data);

    fprintf (stderr, 
	     "Done.  Embedded %lld/%u bytes out of"
	     " a total possible %u bytes.\n"
	     "Encoding rate: 1/%lld\n",
	     bytes_embedded,
	     (unsigned int)mesg_data->sz, 
	     (unsigned int)num_bits/8,
	     code_sz / (num_bits ? num_bits/8 : INT_MAX)
	     );

    /* 
     * cleanup
     */
  out:
    bzero (password, _PASSWORD_LEN);
    if (host_data) free (host_data);
    if (mesg_data) free (mesg_data);
    host_data = NULL;
    mesg_data = NULL;
    while (code_sections)
    {
	tmp_sections = code_sections->next;
	free (code_sections);
	code_sections = tmp_sections;
    }
    disassemble_cleanup ();
    return 0;
}

