/*
 * $Id: hdn_stats.c,v 1.7 2003/02/04 06:54:13 xvr Exp $
 * Created: 08/21/2002
 *
 * xvr (c) 2002-2003
 * xvr@crazyboy.com
 */

#include "hdn_stats.h"

void hdn_stats_embeddable_bits (struct hdn_data *data, unsigned int *num)
{
    long long int curr_pos = 0;
    struct instr insn;

    while (curr_pos < data->sz)
    {
	bzero (&insn, sizeof insn);
	
	disassemble_address (data->content + curr_pos, &insn);

	if ((curr_pos + insn.size) > data->sz) return;

	(*num) += hdn_is_valid_insn (&insn, data->content + curr_pos);
	
	if (insn.size) curr_pos += insn.size;
	else           curr_pos++; //unknown insn, move on
    }
}

static void hdn_usage (void)
{
    fprintf (stderr, 
	     "Usage:\n"
	     "\thydan-stats file\n\n"
	     "Takes in a file, and prints out its statistical profile\n");
    exit (1);
	     
}

/*
 * Returns the number of bits that can be embedded given n functions.
 * Rationale: n functions can be ordered in n! different ways.  So, we
 * can embed log2(n!) bits.
 */
static int _num_bits (long long int n)
{
    double l = 0;

    /*
     * calculate log(n!) iteratively, since n! can be a prohibitively
     * large number and i don't want to use bignum...
     * Reminder for the high-school challenged: 
     *    log (a * b) == log (a) + log (b)
     * For an approximate solution, there is also:
     *    n! =~ sqrt (2*Pi*n) * (n/e)^n
     */
    while (--n > 0)
	l += hdn_log2 (n);

    return l;
}

static long long int count_fns (char *argv0)
{
    struct hdn_data *host_data = NULL;
    struct hdn_host_section *code_sections = NULL, *tmp_sections;
    long long int fn_count = 0;
	

    /*
     * read in application data
     */
    if (!(host_data = hdn_fileread (argv0)))
	goto out;

    /*
     * get the text segment
     */
    if (!hdn_is_valid_exe (host_data->content))
	goto out;

    if (!(code_sections = hdn_get_code_locs (host_data->content)))
    {
	fprintf(stderr, "Error extracting .text segment from host file\n");
	goto out;
    }

    /*
     * count the number of functions in each code section
     */
    tmp_sections = code_sections;
    while (tmp_sections)
    {
	int pos = 0;
	struct instr insn;
	char fn_start[5] = { 0x55, 0x89, 0xe5, 0x83, 0xec};
	
	while (pos < tmp_sections->data.sz)
	{
	    bzero (&insn, sizeof insn);
		
	    disassemble_address (tmp_sections->data.content + pos, &insn);

	    if ((pos + insn.size) > tmp_sections->data.sz) 
		break;

	    //valid instruction
	    if (insn.size)
	    {
		if (!memcmp (tmp_sections->data.content + pos, fn_start, 5))
		    fn_count++;
		
		pos += insn.size;
	    }
	    else pos++; //unknown insn, move on
	}
	
	tmp_sections = tmp_sections->next;
    }
	
    /* 
     * cleanup
     */
  out:
    if (host_data) 
    {
	free (host_data);
	host_data = NULL;
    }

    while (code_sections)
    {
	tmp_sections = code_sections->next;
	free (code_sections);
	code_sections = tmp_sections;
    }

    return fn_count;
}

struct insns
{
    char **desc;
    int  *occ;
    int  *nocc; //negative occurrances

    struct insns *next;
};

struct insns *add_insn (struct insns **insn, char *desc, int bits)
{
    struct insns *temp;

    if (!desc) return NULL;

    /* first time */
    if (!*insn)
    {
	(*insn) = malloc (sizeof (struct insns));
	bzero (*insn, sizeof (struct insns));

	/* allocate 2^(bits+1) possible instructions */
	(*insn)->desc = malloc (sizeof (char *) * (pow (2, bits + 1) + 1));
	bzero ((*insn)->desc,   sizeof (char *) * (pow (2, bits + 1) + 1));
	(*insn)->desc[0] = strdup (desc);

	(*insn)->occ  = malloc (sizeof (int) * pow (2, bits+1));
	bzero ((*insn)->occ,    sizeof (int) * pow (2, bits+1));
	(*insn)->occ[(int)pow (2, bits + 1) - 1] = -1; //terminator

	(*insn)->nocc  = malloc (sizeof (int) * pow (2, bits+1));
	bzero ((*insn)->nocc,    sizeof (int) * pow (2, bits+1));
	(*insn)->nocc[(int)pow (2, bits + 1) - 1] = -1; //terminator

	return (*insn);
    }

    /* find last item on the list */
    temp = *insn;
    while (temp->next)
	temp = temp->next;
	
    /* add the item to the list */
    temp->next = malloc (sizeof (struct insns));
    bzero (temp->next, sizeof (struct insns));

    /* allocate 2^(bits+1) possible instructions */
    temp->next->desc = malloc (sizeof (char *) * (pow (2, bits + 1) + 1));
    bzero (temp->next->desc,   sizeof (char *) * (pow (2, bits + 1) + 1));
    temp->next->desc[0] = strdup (desc);

    temp->next->occ  = malloc (sizeof (int) * pow (2, bits+1));
    bzero (temp->next->occ,    sizeof (int) * pow (2, bits+1));
    temp->next->occ[(int)pow (2, bits + 1) - 1] = -1; //terminator

    temp->next->nocc  = malloc (sizeof (int) * pow (2, bits+1));
    bzero (temp->next->nocc,    sizeof (int) * pow (2, bits+1));
    temp->next->nocc[(int)pow (2, bits + 1) - 1] = -1; //terminator

    return temp->next;
}

struct insns *find_insn (struct insns *insn, char *desc)
{
    while (insn)
    {
	if (!strcmp (insn->desc[0], desc)) 
	    return insn;

	insn = insn->next;
    }

    //not found
    return NULL;
}

void do_stats (struct hdn_data *data, struct insns **insns)
{
    long long int curr_pos = 0;
    int bits;
    struct instr insn;

    while (curr_pos < data->sz)
    {
	bzero (&insn, sizeof insn);
	disassemble_address (data->content + curr_pos, &insn);

	/* done */
	if ((curr_pos + insn.size) > data->sz) return;

	
	if ((bits = hdn_is_valid_insn (&insn, data->content + curr_pos)))
	{
	    char *class_desc, *insn_desc;
	    int val;
	    struct insns *curr_insn;

	    class_desc = 
		hdn_insn_desc (&insn, data->content + curr_pos, &insn_desc);

	    /* do we already have it somewhere */
	    if (!(curr_insn = find_insn (*insns, class_desc)))
	    {
		/* add it */
		curr_insn = add_insn (insns, class_desc, bits);
	    }

	    val = hdn_insn_val (&insn, data->content + curr_pos, NULL);

	    /* inc the occurrence of this value */
	    curr_insn->occ[val] += bits;
	    curr_insn->nocc[val] += bits * hdn_is_neg_insn (&insn, 
							    data->content + 
							    curr_pos);
	    curr_insn->desc[val+1] = insn_desc;
	}
	
	if (insn.size) curr_pos += insn.size;
	else           curr_pos++; //unknown insn, move on
    }
}

int hdn_stats_main (int argc, char **argv)
{
    struct insns *insns = NULL, *tmp_insns;
    struct hdn_data *host_data = NULL;
    struct hdn_host_section *code_sections = NULL, *tmp_sections;

    long long int code_sz = 0, total_code_sz = 0;
    long long int num_fns = 0, total_num_fns = 0;
    unsigned int num_bits, num_fns_bits;
    unsigned int total_num_bits = 0, total_num_fns_bits = 0;
    long long int num_files = 0;

    /*
     * parse that shit
     */
    if (argc < 2)
	hdn_usage ();

    disassemble_init (0, INTEL_SYNTAX);

    argc--;
    while (argc)
    {
	argv++;
	argc--;

	/*
	 * init
	 */
	code_sz = num_fns = num_bits = 0;

	/*
	 * read in the app
	 */
	if (!(host_data = hdn_fileread (*argv)))
	    goto out;

	/*
	 * get the text segment
	 */
	if (!hdn_is_valid_exe (host_data->content))
	    goto out;
	
	if (!(code_sections = hdn_get_code_locs (host_data->content)))
	{
	    fprintf(stderr, "Error extracting .text segment from host file\n");
	    goto out;
	}

	/*
	 * count the number of embeddable instructions in all of the code
	 * sections
	 */
	tmp_sections = code_sections;
	while (tmp_sections)
	{
	    code_sz += tmp_sections->data.sz;
	    
	    hdn_stats_embeddable_bits (&tmp_sections->data, &num_bits);
	    do_stats (&tmp_sections->data, &insns);
	    tmp_sections = tmp_sections->next;
	}

	/*
	 * count the number of functions
	 */
	num_fns = count_fns (*argv);
	num_fns_bits = _num_bits (num_fns);

	/*
	 * stats proper
	 */
	total_num_bits     += num_bits;
	total_code_sz      += code_sz;       
	total_num_fns      += num_fns;
	total_num_fns_bits += num_fns_bits;
	num_files++;

	printf ("File                  : %s\n", *argv);
	printf ("Code size             : %lld bytes\n", code_sz);
	printf ("Embeddeable insns     : %u bytes\n", num_bits/8);
	printf ("Number of functions   : %lld (%d bytes)\n", 
		num_fns, num_fns_bits/8);
	printf ("-------> Encoding Rate: 1/%lld (1/%lld with fns)\n\n",
		code_sz / (num_bits/8 ? 
			   num_bits/8 : INT_MAX),
		code_sz / ((num_bits + num_fns_bits)/8 ? 
			   (num_bits + num_fns_bits)/8 : INT_MAX));

	printf ("Total embeddeable insns: %u bytes\n", total_num_bits/8);
	printf ("Total number of fns    : %lld (%d bytes/file, %d bytes/total)\n",
		total_num_fns, 
		total_num_fns_bits/8, _num_bits (total_num_fns)/8);
	printf ("Total encoding rate    : 1/%lld (with fns: 1/%lld/file - 1/%lld/total)\n\n",
		total_code_sz / (total_num_bits/8 ?
				 total_num_bits/8 : INT_MAX),
		total_code_sz / ((total_num_bits + total_num_fns_bits)/8 ?
				 (total_num_bits + total_num_fns_bits)/8 
				 : INT_MAX),
		total_code_sz / ((total_num_bits + _num_bits (total_num_fns))/8 ?
				 (total_num_bits + _num_bits (total_num_fns))/8
				 : INT_MAX));

	tmp_insns = insns;
	while (tmp_insns)
	{
	    int i;
	    
	    printf ("%s:\n", tmp_insns->desc[0]);
	    for (i = 0; tmp_insns->occ[i] != -1; i++)
	    {
		if (tmp_insns->occ[i])
		    printf ("\t%s: %d\n", tmp_insns->desc[i+1], 
			    tmp_insns->occ[i]);
		if (tmp_insns->nocc[i])
		    printf ("\t\tof which are in the negative form: %d\n",
			    tmp_insns->nocc[i]);
	    }
	    
	    tmp_insns = tmp_insns->next;
	}

	/* 
	 * cleanup
	 */
      out:
	if (host_data) 
	{
	    free (host_data);
	    host_data = NULL;
	}
	
	while (code_sections)
	{
	    tmp_sections = code_sections->next;
	    free (code_sections);
	    code_sections = tmp_sections;
	    tmp_sections = NULL;
	}	
    }

    printf ("num files: %lld\n", num_files);

    disassemble_cleanup ();
    return 0;
}
